--[[
**********************************************************************
MagicDKP - a raid loot and DKP tracker.
**********************************************************************
This file is part of MagicDKP, a World of Warcraft Addon

MagicDKP is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MagicDKP is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MagicDKP.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************************
]]

MagicDKP = LibStub("AceAddon-3.0"):NewAddon("MagicDKP", "AceConsole-3.0",
					    "AceEvent-3.0", "AceTimer-3.0", "LibLogger-1.0")
MagicDKP.MAJOR_VERSION = "MagicDKP-1.0"
MagicDKP.MINOR_VERSION = tonumber('136') or tonumber(("$Revision: 130 $"):match("%d+"))

local MagicComm = LibStub("MagicComm-1.0")
local C = LibStub("AceConfigDialog-3.0")
local GUI = LibStub("AceGUI-3.0")
local BabbleBoss = LibStub("LibBabble-Boss-3.0") 
local LDBIcon = LibStub("LibDBIcon-1.0", true)
local LDB = LibStub("LibDataBroker-1.1", true)
local BL = BabbleBoss:GetUnstrictLookupTable()
local GetItemCount, GetItemInfo, match = GetItemCount, GetItemInfo, string.match
local raidEntered
BabbleBoss = nil

MagicDKPDB = {}

local mod = MagicDKP
local L = LibStub("AceLocale-3.0"):GetLocale("MagicDKP", false)

local UnitGUID = UnitGUID
local sub = string.sub
local UnitClassification = UnitClassification
local table_insert = table.insert
local loggingDeaths = false
local fmt = string.format

local defaultOptions = {
   profile = {
      lootAwarded = "[player] awarded [item] for [dkp] dkp.",
      lootNotification = true,
      logLevel = mod.logLevels.INFO,
      raidsize = "25",
      trackwipe = false,
      wipepercentage = 50,
      standbyattendee = false,
      eventnameaszone = false,
      trackonly = false,
      logPlayerEventData = true,
      reportraid = true,
      reportstandby = true,
      enabled = true,
      detectAlts = true,
      standbyTracking = true,
      minimapIcon = {},
      lootEventListen = true,
      maxraidage = 1,
      xloothook = false,
      external = true,
      modules = {}
   }
}

local db, zones, currentZone, currentRaid, currentEvent, currentBoss
local playerName = UnitName("player")

local networkData = {}

function mod:SetNetworkData(cmd, data, misc1, misc2, misc3, misc4)
   networkData.cmd   = cmd
   networkData.data  = data
   networkData.misc1 = misc1
   networkData.misc2 = misc2
   networkData.misc3 = misc3
   networkData.misc4 = misc4
end
      

-- Dummy debug ftw.
function mod:get()
   return {
      zone = currentZone,
      raid = currentRaid,
      event = currentEvent
   }
end

local function UpdateUpvalues()
   db = mod.db.profile         -- profile data
   zones = MagicDKPDB.zones     -- raid zones / boss mobs
   items = MagicDKPDB.items     -- item database
   events = MagicDKPDB.events   -- custom events
   raids = MagicDKPDB.raids     -- raids
   players = MagicDKPDB.players -- player db
end

function mod:OnInitialize()
   self.db = LibStub("AceDB-3.0"):New("MagicDKPConfigDB", defaultOptions, "Default")
   self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
   self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
   self.db.RegisterCallback(self, "OnProfileDeleted","OnProfileChanged")
   self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")

   MagicDKPDB.zones    = MagicDKPDB.zones    or {}
   MagicDKPDB.items    = MagicDKPDB.items    or mod.defaultItems
   MagicDKPDB.events   = MagicDKPDB.events   or mod.defaultEvents  
   MagicDKPDB.raids    = MagicDKPDB.raids    or {}
   MagicDKPDB.players  = MagicDKPDB.players  or {}

--   mod.defaultItems  = nil 
   mod.defaultEvents = nil 
   
   UpdateUpvalues()

   self:SetLogLevel(db.logLevel)
   self:AddDefaultZones() -- Populate MagicDKPDB with the default zones
   self:BuildConfiguration()   
   if LDB then
      self.ldb =
	 LDB:NewDataObject("MagicDKP",
			   {
			      type =  "launcher", 
			      label = "MagicDKP",
			      icon = "Interface\\AddOns\\MagicDKP\\dkp.tga",
			      tooltiptext = L["|cffffffffMagicDKP|r\n|cffffff00Left click|r to toggle the configuration screen.\n"],
			      OnClick = function(clickedframe, button)
					   if button == "LeftButton" then
					      mod:ToggleConfigDialog()
					   end
					end,
			   })
      if LDBIcon then
	 LDBIcon:Register("MagicDKP", self.ldb, mod.db.profile.minimapIcon)
      end
   end
end



local currentNoteItem = {}
local noteItemQueue = {}
local itemNoteText = L["Enter value for %s, looted by %s."]
local itemLootDialog

local function CheckQueuedItem()
   local count =  #noteItemQueue
   local item = count > 0 and noteItemQueue[count]
   if item then
      mod:AddItemDialog(item.looter, item.itemid, item.value, item.link, item.count)
      noteItemQueue[count] = nil
   else
      currentNoteItem.itemid = nil
      currentNoteItem.looter = nil
      currentNoteItem.value = nil
      currentNoteItem.count = nil
      itemLootDialog:Hide()
   end
end

local function QueueItem(looter, itemid, value, link, count)
   local queueSize = #noteItemQueue
   if queueSize > 0 or currentNoteItem.itemid then
      noteItemQueue[#noteItemQueue+1] = {
	 looter = looter,
	 itemid = itemid,
	 value = value or 0,
	 link = link,
	 count = count 
      }
   else
      mod:AddItemDialog(looter, itemid, value or 0, link, count)
   end
   
end

-- Call external addons adjusting the dkp of a player
-- Currently used only when spending DKP.
function mod:AdjustExternalDKP(looter, value, pool)
   local dkpOwner = looter
   if player[looter] and player[looter].main then
      -- attribute dkp to the main, not the alt
      dkpOwner = player[looter].main
   end
   if ShowDKPAdjustPlayerDKP then
      ShowDKPAdjustPlayerDKP(dkpOwner, value)
   end
   local DB = mod:Database()
   if DB then
      DB:AdjustSpentDKP(dkpOwner, value, pool)
      mod:SendDKPBalance(looter)
      mod:SendDKPBalance(dkpOwner)
   end
end

function mod:GetCurrentEvent()
   if currentEvent then return currentEvent end

   local key,name = self:GetZoneName(currentZone.name)
   if not currentRaid.events[name] then
      currentRaid.events[name] = {
	 attendees = { },
	 eventTime = self:GetGameTimeAsUnix(), -- set event time to time of first looted item
	 items = {}
      }
   end
   local attendees = currentRaid.events[name].attendees
   for i = 1, GetNumRaidMembers() do  
      local name = GetRaidRosterInfo(i)
      attendees[GetRaidRosterInfo(i)] = true
   end
   return currentRaid.events[name]
end

function mod:AddItem(looter, itemid, value, count)
   if self.hasTrace then
      self:trace("Adding item %sx%d for %s at value %d.", items[itemid].itemInfo[2], count, looter, value)
   end
   if currentRaid then
      mod:SendLootNote(looter, items[itemid].itemInfo[2], value)
      local newitem = {
	 itemid = itemid,
	 looter = looter,
	 value = value,
	 count = count,
	 multidkp = mod:GetMultiPool(itemid)
      }
      mod:AdjustExternalDKP(looter, value, newitem.multidkp)
      
      if currentEvent then
	 currentEvent.items[#currentEvent.items+1] = newitem
      else
	 local event = mod:GetCurrentEvent()
	 event.attendees[looter] = true
	 event.items[#event.items+1] = newitem
      end

      for id,data in pairs(raids) do
	 if data == currentRaid then
	    self:AddRaidConfig(id, build)
	    break;
	 end
      end
      mod:ScheduleRaidChangeData(currentRaid)
      self:NotifyChange()
   end  
end

function mod:AddItemDialog(looter, itemid, value, link, count)
   currentNoteItem.itemid = itemid
   currentNoteItem.looter = looter
   currentNoteItem.value = value
   currentNoteItem.count = count
   
   if not itemLootDialog then
      itemLootDialog = GUI:Create("MDKP_Dialog")
      itemLootDialog:SetTitle(L["Add Item"])
      itemLootDialog:SetLayout("Flow")

      itemLootDialog.ok = GUI:Create("Button")
      itemLootDialog.cancel = GUI:Create("Button")
      itemLootDialog.bank = GUI:Create("Button")
      itemLootDialog.de = GUI:Create("Button")

      itemLootDialog:AddChild(itemLootDialog.ok)
      itemLootDialog:AddChild(itemLootDialog.cancel)
      itemLootDialog:AddChild(itemLootDialog.bank)
      itemLootDialog:AddChild(itemLootDialog.de)

      itemLootDialog.ok:SetText(L["Ok"])
      itemLootDialog.cancel:SetText(L["Ignore"])
      itemLootDialog.bank:SetText(L["Bank"])
      itemLootDialog.de:SetText(L["Disenchant"])

      local onAccept = function()
			  local val = itemLootDialog:GetValue()
			  mod:AddItem(currentNoteItem.looter, currentNoteItem.itemid, val, currentNoteItem.count)
			  items[currentNoteItem.itemid].value = val
			  CheckQueuedItem()
		       end
      
      itemLootDialog:SetCallback("OnAccept", onAccept)
      itemLootDialog.ok:SetCallback("OnClick", onAccept)
      itemLootDialog.cancel:SetCallback("OnClick",
				    function()
				       items[currentNoteItem.itemid].ignore = true
				       items[currentNoteItem.itemid].value = nil
				       items[currentNoteItem.itemid].bank = nil
				       CheckQueuedItem()
				    end
				 )
      itemLootDialog.bank:SetCallback("OnClick",
				    function()
				       items[currentNoteItem.itemid].bank = true
				       items[currentNoteItem.itemid].value = nil
				       mod:AddItem("bank", currentNoteItem.itemid, 0, currentNoteItem.count)
				       CheckQueuedItem()
				    end
				 )
      itemLootDialog.de:SetCallback("OnClick",
				    function()
				       mod:AddItem("disenchant", currentNoteItem.itemid, 0, currentNoteItem.count)
				       CheckQueuedItem()
				    end
				 )
      itemLootDialog.ok:SetWidth(80)
      itemLootDialog.cancel:SetWidth(80)
      itemLootDialog.bank:SetWidth(80)
      itemLootDialog.de:SetWidth(125)
   end

   
   itemLootDialog:SetText(itemNoteText:format(link, looter ))
   itemLootDialog:SetValue(value)

   itemLootDialog:Show()
   return itemLootDialog
   
end


function mod:OnEnable()
   if not db.enabled then return end
   
   local popup = _G.StaticPopupDialogs
   if type(popup) ~= "table" then
      popup = {}
   end
   if type(popup["MDKPAttendance"]) ~= "table" then
      popup["MDKPAttendance"] = {
	 text = "",
	 button1 = TEXT(YES),
	 button2 = TEXT(NO),
	 timeout = 0,
	 whileDead = 1,
	 hideOnEscape = 0,
	 OnAccept = function() self:TakeAttendance() end
      }
   end
   
   if type(popup["MDKPDuplicate"]) ~= "table" then
      popup["MDKPDuplicate"] = {
	 text = "",
	 button1 = TEXT(YES),
	 button2 = TEXT(NO),
	 timeout = 0,
	 whileDead = 1,
	 hideOnEscape = 0,
	 OnAccept = function() self:QueueDuplicateItem() end,
	 OnClose = function() self:ResetDuplicateItem() end
      }
   end
   
   MagicComm:RegisterListener(self, "MD")
   self:RegisterEvent("RAID_ROSTER_UPDATE", "StartOrEndRaid")
   self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
   self:ZONE_CHANGED_NEW_AREA()
   self:RegisterEvent("GUILD_ROSTER_UPDATE")
   GuildRoster()
   self:LoadXLootHook()
end

function mod:LoadXLootHook()
   if db.xloothook and not IsAddOnLoaded("MagicDKP_XLootHook") and IsAddOnLoaded("XLootMaster") then
      LoadAddOn("MagicDKP_XLootHook")
   end
end

function mod:ParseAltNote(note)
   if not note then return end
   local alt =  match(note, "Alt:(.*)")
   if alt then return strtrim(alt) end
end

local badData = { ['<AFK>'] = true, ['<DND>'] = true }
function mod:UpdatePlayerInfo(name, rank, level, class, race, sex, note, officernote)
   if badData[sex] then sex = nil end -- wtf
   if badData[race] then race = nil end -- ugly hack
   local hash = players[name] or {}
   hash.rank = rank or hash.rank or "Unknown"
   hash.level = level or hash.level or 0
   hash.class = class or hash.class or "Unknown"
   hash.race = race or hash.race or "Unknown"
   hash.sex = sex or hash.sex or 0
   players[name] = hash
   if db.detectAlts then
      local alt
      if note then
	 alt = self:ParseAltNote(note)
      end
      if not alt and officernote then
	 alt = self:ParseAltNote(officernote)
      end
      if alt then
	 if hash.main ~= alt then
	    if not players[alt] then
	       if self.hasWarn then self:warn("Found main character %s for %s, but main character not found.", alt, name) end
	    else
	       if self.hasInfo then self:info("Found main character %s for %s.", alt, name) end	       
	       hash.main = alt
	    end
	 end
	 if players[alt] then
	    if players[alt].main then
	       if self.hasWarn then self:warn("Removing main character %s from detected alt %s", players[alt].main, alt) end
	       players[alt].main = nil
	       mod:AddPlayerConfig(alt)
	    end
	 else
	    if self.hasSpam then self:spam("Detected main character %s (alt of %s) not found in player database.", alt, name) end
	 end
      elseif db.clearAlt and hash.main then
	 if self.hasWarn then self:warn("Removing main character %s from %s due to missing guild alt note.", hash.main, name) end
	 hash.main = nil
	 mod:AddPlayerConfig(name)
      end
   end
   mod:AddPlayerConfig(name)
   mod:NotifyChange()
end

function mod:GUILD_ROSTER_UPDATE()
   self:UnregisterEvent("GUILD_ROSTER_UPDATE") 
   local numGuildMembers = GetNumGuildMembers(true)
   for i = 1, numGuildMembers do
      local name, rank, rankIndex, level, class, zone, note, officernote, _, _, class = GetGuildRosterInfo(i)
      self:UpdatePlayerInfo(name, rank, level, class, nil, nil, note, officernote)
   end
end

function mod:TakeStandbyAttendance(event)
   local n = GetNumGuildMembers()
   if not n or n < 1 then return end

   self:UnregisterEvent("GUILD_ROSTER_UPDATE") 

   if db.reportstandby then
      mod:SendEventNotice(event.name, "GUILD")
   end

   local numGuildMembers = GetNumGuildMembers(true)
   for i = 1, numGuildMembers do
      local name, rank, rankIndex, level, class, zone, note, officernote, _, _, class = GetGuildRosterInfo(i)
      if not event.attendees[name] and not event.standby[name] and
	 mod:IsRaiderRank(rankIndex) then
	 self:UpdatePlayerInfo(name, rank, level, class, nil, nil, note, officernote)
	 mod:SetNetworkData("STANDBYCHECK", event.name, #raids)
	 self:SendUrgentMessage("WHISPER", name)
	 if self.hasTrace then self:trace("Sending attendance message to %s.", name) end
      end
   end
end

function mod:IsRaiderRank(rankIndex)
   return not db.ranks or db.ranks[rankIndex+1]
end

do
   local eventTimers = {}
   function mod:ScheduleRaidChangeData(raid)
      if mod:IsLootMaster() then
	 if eventTimers[raid.startTime] then mod:CancelTimer(eventTimers[raid.startTime], true) end
	 eventTimers[raid.startTime] = mod:ScheduleTimer("SendEventSyncData", 360, raid)
      end
   end

   function mod:SendEventSyncData(raid)
      if eventTimers[raid.startTime] then mod:CancelTimer(eventTimers[raid.startTime], true) end
      eventTimers[raid.startTime] = nil
      if mod:IsLootMaster() then
	 mod:SendRaidSyncData(raid)
      end
   end
end

function mod:OnStandbyResponse(player, eventName, raidid)
   local raid
   if raidid and raids[raidid] then
      raid = raids[raidid]
   else
      raid = raids[#raids]
   end
   local event = raid.events[eventName]
   if event then
      if not event.attendees[player] then
	 event.standby[player] = true
	 if self.hasInfo then self:info("Adding %s as standby for event %s.", player, eventName) end
	 mod:ScheduleRaidChangeData(raid)
      else
	 if self.hasTrace then self:trace("Not adding %s to standby - already attending.", player) end
      end
   end
end

-- Returns true if we're the master looter @ master loot mode, or raid leader otherwise
-- I.e it means we should collect data and be the authorative source
function mod:IsLootMaster()
   local method, mlparty = GetLootMethod()
   return (method == "master" and mlparty == 0) or (method ~= "master" and IsRaidLeader()) or GetNumRaidMembers() == 0
end

function mod:RequestInitialSyncData()
   mod:SetNetworkData("DKPSYNC", "SENDRAID")
   mod:SendBulkMessage()
end

function mod:SendRaidSyncData(raid)
   if not mod:IsLootMaster() then return end
   raid = raid or currentRaid
   if raid then
      mod:SetNetworkData("DKPSYNC", "ADDRAID", raid)
      mod:SendBulkMessage("RAID")
   end
end

function mod:SendDeleteRaid(raidid)
   if not mod:IsLootMaster() then return end
   mod:SetNetworkData("DKPSYNC", "DELRAID", raidid)
   mod:SendBulkMessage("RAID")
end

function mod:DelSyncRaid(raidid)
   if raids[#raids] and raids[#raids].startTime == raidid then
      mod:EndRaid()
   end
end

function mod:AddSyncRaid(data)
   -- Don't add if you're the loot meister
   if mod:IsLootMaster() then return end
   -- Add all items found in the new raid
   for _, event in pairs(data.events) do
      if event.items then
	 for _, item in ipairs(event.items) do
	    if not items[item.itemid] then
	       self:UpdateItem(item.itemid)
	    end
	 end
      end
   end

   if currentRaid then mod:EndRaid(true) end

   local replacedRaid
   for id,raid in pairs(raids) do
      if raid.startTime == data.startTime then
	 -- Found a match, just replace the old data
	 raids[id] = data
	 if not data.endTime then currentRaid = data end
	 replacedRaid = true
	 break
      end
   end
   
   if not replacedRaid then
      raids[#raids +1] = data
      if not data.endTime then currentRaid = data end
   end
   mod:RefreshRaidConfig()
   mod:NotifyChange()
end

local syncHandlers = {
   SENDRAID = mod.SendRaidSyncData,
   ADDRAID = mod.AddSyncRaid,
}

function mod:OnDKPSyncRequest(command, ...)
   if syncHandlers[command] then
      syncHandlers[command](mod, ...)
   end
end


function mod:OnDisable()
   self:UnregisterEvent("CHAT_MSG_LOOT")
   self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
   self:UnregisterEvent("GUILD_ROSTER")
   self:UnregisterEvent("PLAYER_REGEN_DISABLED")
   self:UnregisterEvent("PLAYER_TARGET_CHANGED")
   self:UnregisterEvent("RAID_ROSTER_UPDATE")
   self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")   
   self:UnregisterEvent("ZONE_CHANGED_NEW_AREA")
   MagicComm:UnregisterListener(self, "MD")
end

local function UnitIsEligable (unit, uid)
   return UnitClassification(unit) == "worldboss" and (UnitCanAttack("player", unit)  or uid == 24892) -- special case for sathorarr since he's considered friendly when dead
end


local function GUIDToUID(guid)
   local uid = tonumber(sub(guid, 7, 12), 16)
   if uid == 0 then return end
   return uid
end

local function GetUnitID(unit)
   local guid = UnitGUID(unit)
   return guid, GUIDToUID(guid)
end


function mod:ScanTarget(unit)
   if not UnitExists(unit) then return end
   local guid, uid = GetUnitID(unit)

   if not currentZone or not UnitIsEligable(unit, uid) then
      return
   end

   if uid and not currentZone.mobs[uid] then
      self:AddNewBoss(uid, UnitName(unit), currentZone)
   end
   
   if UnitIsDead(unit) then
      -- we found a dead boss, check to see if this guy is killed
      if not currentEvent and currentRaid then
	 local event = currentZone.mobs[uid].event or currentZone.mobs[uid].name
	 local tmpEvent = currentRaid.events[event] 
	 if tmpEvent then
	    if self.hasTrace then self:trace("Found dead boss mob with finished event - activating.") end
	    currentEvent = tmpEvent
	    self:ScheduleTimer("ClearEvent", 120, currentEvent) -- clear it after 120 seconds
	 elseif not currentZone.mobs[uid].ignore then
	    if self.hasTrace then self:trace("Possibly logging event for [mouseover] %s: %s.", event, tostring(uid)) end
	    self:LogRaidEvent(uid, currentZone.mobs[uid])
	 end
      end
   elseif not loggingDeaths then
      if self.hasTrace then self:trace("Detected boss mob, enabling combat event log filtering.") end 
      self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED") -- eligable mob found, enable death tracking
      loggingDeaths = true
   end
end

-- From when friendly mobs were added
local blacklist = {
   24895, -- Madrigosa / SWP
   17948, -- Tyrande / Hyjal
   17773, -- Jaina / Hyjal
   17852, -- Thrall / Hyjal
   24891, -- Human Kalecgos, SWP
   23197, -- Maiev Shadowsong, BT
}
   
function mod:AddDefaultZones()
   if not mod.defaultZones then
      return
   end
   local key
   for zone, zonedata in pairs(mod.defaultZones) do
      key = self:MakeKey(zone)
      if not zones[key] then
	 zones[key] = {
	    name = zone,
	    size = zonedata.size,
	    mobs = { }
	 }
      end
      zone = zones[key].mobs
      for _, mob in ipairs(blacklist) do
	 zone[mob]  = nil -- remove these 
      end
      for mob, data in pairs(zonedata.mobs) do
	 if not zone[mob] then
	    zone[mob] = {}
	    for key,val in pairs(data) do
	       zone[mob][key] = val
	    end
	 end
      end
   end
end

function mod:ZONE_CHANGED_NEW_AREA()
   local zone,name = self:GetZoneName()
   currentZone = nil
   if zone == nil or zone == "" then
      self:ScheduleTimer("ZONE_CHANGED_NEW_AREA",5)
   else
      local inInstance, type = IsInInstance()
	 
      if type ~= "raid" then  -- FIXME
	 zone = "WorldBoss"
	 name = "World Boss"
      end
      if not zones[zone] then
	 zones[zone] = {
	    name = name, 
	    mobs = {}
	 }
	 self:BuildConfiguration()
	 self:NotifyChange()
      end

      currentZone = zones[zone]
      if not currentRaid then
	 self:StartOrEndRaid() -- make sure we have a raid if we should
      end
      if currentRaid and not currentRaid.zone and type == "raid" then
	 currentRaid.zone = name
	 self:NotifyChange()
	 if self.hasTrace then self:trace("Set %s to the zone of the current raid.", name) end
      end
   end
end

local linkpat = '|c........|Hitem:(%%d+):.-|h|r'
local globalpatterns = {
   "LOOT_ITEM",
   "LOOT_ITEM_SELF",
   "LOOT_ITEM_MULTIPLE",
   "LOOT_ITEM_SELF_MULTIPLE",
}

local patterns = {}
for _, global in ipairs(globalpatterns) do
   local pattern = _G[global]
   if sub(pattern, 1,2) == "%s" then
      pattern = "([^ ]+)"..sub(pattern, 3)
   end
   patterns[global] = gsub(gsub(pattern, "%%d", "(%%d+)"), "%%s", linkpat)
end

function mod:CHAT_MSG_LOOT(event, msg)
   local looter, itemid, count

   if self.hasSpam then self:spam("Parsing loot event: %s", msg) end
   looter, itemid, count = match(msg, patterns["LOOT_ITEM_MULTIPLE"])

   if looter and itemid and count then
      mod:HandleLoot(looter, tonumber(itemid), tonumber(count))
      return
   end
   
   itemid, count  = match(msg, patterns["LOOT_ITEM_SELF_MULTIPLE"])
   if itemid and count then
      mod:HandleLoot(playerName, tonumber(itemid), tonumber(count))
      return
   end

   looter, itemid = match(msg, patterns["LOOT_ITEM"])
   if looter and itemid then
      mod:HandleLoot(looter, tonumber(itemid), 1)
      return
   end

   itemid = match(msg, patterns["LOOT_ITEM_SELF"])
   if itemid then
      mod:HandleLoot(playerName, tonumber(itemid), 1)
      return
   end
end

function mod:AddItemManually(_,cmd)
   if not db.enabled then
      if self.error then self:error("The addon isn't enabled.") end
      return
   end
   local looter,itemid = match(cmd, "([^ ]+) .*|Hitem:(%d+):")
   if not looter or not itemid then
      if self.error then self:error("Invalid syntax. Expected: <looter> [itemlink], got '%s'", cmd) end
      return
   end
   self:HandleLoot(looter, tonumber(itemid), 1, true)
end

function mod:StartDKPBid(_,cmd)
   if not db.enabled then
      if self.error then self:error("The addon isn't enabled.") end
      return
   end
   local itemid = mod:LinkToItemID(cmd)
   if not itemid then
      if self.error then self:error("Invalid syntax. Expected: [itemlink], got '%s'", cmd) end
      return
   end
   if not mod.modules.Bidder then
      LoadAddOn("MagicDKP_Bidder")
   end
   if not mod.modules.Bidder then
      if self.error then self:error("Cannot do DKP Bid, MagicDKP Bidder not found") end
      return
   end
   mod.modules.Bidder:StartNewBid(select(2, GetItemInfo(itemid)))
end


local duplicateItem = {}
function mod:QueueDuplicateItem()
   if duplicateItem.itemid then
      if self.hasSpam then self:spam("adding %s ", duplicateItem.link) end      
      local itemData = items[duplicateItem.itemid]
      if itemData.bank then
	 self:AddItem("bank", duplicateItem.itemid, 0, duplicateItem.count)
      elseif itemData.fixed or db.trackonly then
	 self:AddItem(duplicateItem.looter, duplicateItem.itemid, (itemData.value or 0)*duplicateItem.count, duplicateItem.count)
      else 
	 QueueItem(duplicateItem.looter, duplicateItem.itemid, (itemData.value or 0), duplicateItem.link, duplicateItem.count)
      end
   end
   mod:ResetDuplicateItem()
end

function mod:ResetDuplicateItem()
   duplicateItem.looter = nil
   duplicateItem.itemid = nil
   duplicateItem.link = nil
   duplicateItem.count = nil
end

local lastItem, lastLooter, lastCount, lastTime
local duplicateItemText = L["%s was already awarded %s in this event. Do you want to add another entry"]

function mod:HandleLoot(looter, itemid, count, force, external, isBank, isDE, defaultDKP)
   if (not db.xloothook or not db.external) and external then
      -- external loot hook disabled, ignore events
      return
   end

   local now = time()
   if not currentRaid then
      if not external then
	 self:Print("Unable to add item - no active raid found.")
      end
      return
   end

   if lastItem == itemid and looter == lastLooter
      and count == lastCount and
      lastTime  and (now - lastTime < 15) then
      if self.hasSpam then self:spam("Double loot event detected.") end
      return
   end

   -- dup detection
   lastItem   = itemid
   lastLooter  = looter
   lastCount  = count 
   lastTime   = now
   
   local name,link,rarity = GetItemInfo(itemid)
   if self.hasSpam then self:spam("handling %s", link, " => ", itemid) end

   if not force and rarity < 4 and not items[itemid] then
      if self.hasSpam then self:spam("Ignoring %s [low item quality]", link) end
      return
   end
   
   if not items[itemid] then
      self:UpdateItem(itemid)
   end

   local itemData = items[itemid]

   if itemData.ignore and not force then
      if self.hasSpam then self:spam("Ignoring %s [ignored in database]", link) end
      return
   end

   if defaultDKP and not itemData.fixed then
      itemData.value = defaultDKP
   end

   local event =  mod:GetCurrentEvent()

   if event then
      for id,item in pairs(event.items) do
	 if item.looter == looter and item.itemid == itemid then
	    duplicateItem.looter = looter
	    duplicateItem.itemid = itemid
	    duplicateItem.link = link
	    duplicateItem.count = count
	    
	    StaticPopupDialogs["MDKPDuplicate"].text = duplicateItemText:format(looter, link)
	    StaticPopup_Show("MDKPDuplicate")
	    return
	 end
      end
   end

   if self.hasSpam then self:spam("adding %s ", link) end

   if itemData.bank or isBank then
      self:AddItem("bank", itemid, 0, count)
   elseif itemData.fixed or db.trackonly then
      self:AddItem(looter, itemid, (itemData.value or 0)*count, count)
   elseif isDE then
      self:AddItem("disenchant", itemid, 0, count)
   else 
      QueueItem(looter, itemid, (itemData.value or 0), link, count)
   end
end

do
   local info = {}
   function mod:SendLootNote(player, item, dkp)
      if db.lootNotification then
	 info.player = player
	 info.item = item
	 info.dkp = tostring(dkp)
	 if db.notifyGuild then
	    mod:SendChatMessage(mod:tokenize(db.lootAwarded, info), "GUILD")
	 else
	    mod:SendChatMessage(mod:tokenize(db.lootAwarded, info), "GROUP")
	 end
      end
   end
end

function mod:SendEventNotice(name, dest)
   if dest == "RAID" then
      SendChatMessage(L["MagicDKP: Logging raid attendance for "]..name, dest)
   else
      SendChatMessage(L["MagicDKP: Taking standby DKP for event "]..name, dest)
   end
end

      
function mod:TakeAttendance()
   local groups = 5
   if currentZone and currentZone.size then
      groups = math.floor(currentZone.size / 5)
   end
   if not currentEvent then return end
   currentEvent.eventTime = self:GetGameTimeAsUnix()

   if db.reportraid then
      mod:SendEventNotice(currentEvent.name, "RAID")
   end
   
   for i = 1, GetNumRaidMembers() do  
      local name, _, group, level, _, class = GetRaidRosterInfo(i)
      local id = "raid"..i
      self:UpdatePlayerInfo(name, nil, level == 0 and nil or level, class, select(2, UnitRace(id)), UnitSex(id))
      if group <= groups then
	 currentEvent.attendees[name] = true
	 if self.hasTrace then self:trace("Adding %s to %s.", name, currentEvent.name) end
      else
	 currentEvent.standby[name] = true
	 if self.hasTrace then self:trace("Adding %s as standby to %s.", name, currentEvent.name) end
      end
   end

   mod:ScheduleRaidChangeData(currentRaid)

   if db.standbyTracking then
      self:RegisterEvent("GUILD_ROSTER_UPDATE", "TakeStandbyAttendance", currentEvent)
      GuildRoster()
   end
   
   currentRaid.events[currentEvent.name] = currentEvent
   self:BuildConfiguration()
   self:ScheduleTimer("ClearEvent", 300, currentEvent)
   self:NotifyChange()
end


local raidAttendanceText = L["Record attendance for %s?"]
local eventInProgress

function mod:LogRaidEvent(uid, unitData)
   if not currentRaid then return end -- should never happen
   local event = unitData.event or unitData.name 
   if not eventInProgress or event ~= eventInProgress.name then
      if currentRaid.events[event] then
	 eventInProgress = currentRaid.events[event]
      else
	 if self.hasTrace then self:trace("Creating a new event...") end
	 eventInProgress = {
	    attendees = {},
	    standby = {},
	    items = {},
	    needed = {},
	    name = event,
	    zone = currentZone.name,
	 }
	 for mobuid,data in pairs(currentZone.mobs) do
	    if (data.event or data.name) == event and not data.ignore then
	       eventInProgress.needed[mobuid] = true
	    end
	 end
      end
   end

   eventInProgress.needed[uid] = nil

   if not next(eventInProgress.needed) then 
      eventInProgress.needed = nil -- don't need to save this
      if self.hasTrace then self:trace("Event finished, taking attendance...") end
      if loggingDeaths then
	 loggingDeaths = false
	 -- we're done with this event so no longer need track this
	 self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED") 
      end
      currentEvent = eventInProgress
      eventInProgress = nil
      currentBoss = nil
      if not currentEvent.eventTime then
	 StaticPopupDialogs["MDKPAttendance"].text = raidAttendanceText:format(event)
	 StaticPopup_Show("MDKPAttendance")
      end
   end
end


function mod:ClearEvent(event)
   if event == currentEvent then
      currentEvent = nil
   end
end
function mod:PLAYER_REGEN_DISABLED()
   eventInProgress = nil
   currentBoss = nil
end

function mod:COMBAT_LOG_EVENT_UNFILTERED(_, _, event, _, _, _, guid)
   local uid
   if not currentZone or not currentRaid then
      self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
      return
   end
   if not currentBoss then
      uid = GUIDToUID(guid) 
      if currentZone.mobs[uid] and not currentZone.mobs[uid].isfarm  then
	 -- This means we're engaged with a boss that gives wipe dkp
	 if self.hasTrace then self:trace("Engaged boss: %s", currentZone.mobs[uid].name) end
	 currentBoss = uid
      end
   end
   
   if event ~= "UNIT_DIED" and event ~= "PARTY_KILL" then return end

   uid = uid or GUIDToUID(guid)
   
   if uid and currentZone.mobs[uid] and not currentZone.mobs[uid].ignore then
      if self.hasTrace then self:trace("Possibly logging event for %s: %s.", event, tostring(uid)) end
      self:LogRaidEvent(uid, currentZone.mobs[uid])
   end
end

local isInRaid

-- Get the current raid or create a new one if none is active
function mod:FindOrCreateRaid()
   if not mod:IsLootMaster() then return end
   local raid = raids[#raids]
   local changed = false
   if not raid or raid.endTime then
      raid = {
	 startTime = self:GetGameTimeAsUnix(),
	 events = {
	 },
      }
      raids[#raids+1] = raid
      if self.hasDebug then self:debug("Created new raid.") end
      changed = true
      self:AddRaidConfig(#raids)
   else
      if self.hasDebug then self:debug("Found active raid.") end
   end
   currentRaid = raid
   mod:CheckRaidChanges(true)
   if not currentZone then
      self:ZONE_CHANGED_NEW_AREA()
   end
   if not raid.zone and currentZone and currentZone.name ~= "World Boss" then
      raid.zone = currentZone.name
      changed = true
   end
   if changed then
      self:NotifyChange()
   end
   mod:SendRaidSyncData()
end

function mod:EndRaid(quiet)
   currentRaid = raids[#raids]
   if currentRaid and not currentRaid.endTime  then
      if next(currentRaid.events) then
	 mod:AddEndOfRaidLeaves()
	 currentRaid.endTime = self:GetGameTimeAsUnix()
	 if not quiet and self.hasDebug then self:debug("Ending active raid.") end
	 self:NotifyChange()
	 mod:SendRaidSyncData()
      else
	 -- No events, don't save
	 self:DeleteRaidConfig(#raids)
	 mod:SendDeleteRaid(currentRaid.startTime)
	 raids[#raids] = nil
	 if not quiet and self.hasDebug then self:debug("Deleting active raid (no events).") end
	 self:NotifyChange()
      end
   elseif not quiet and self.hasTrace then
      self:trace("No active raid found.")
   end
   currentEvent = nil
   currentRaid = nil
   currentZone = nil
   self:RegisterEvent("RAID_ROSTER_UPDATE", "StartOrEndRaid")
end

do
   local raidMembers = {}
   local newRaidMembers = {}

   function mod:AddEndOfRaidLeaves()
      for name in pairs(raidMembers) do
	 mod:LogPlayerLeave(name)
	 raidMembers[name] = nil
      end
   end
   function mod:CheckRaidChanges(clear)
      local tmp
      local num = GetNumRaidMembers()
      if clear == true then for name in pairs(raidMembers) do raidMembers[name] = nil end end
      if not currentRaid then return end
      if num == 0 then
	 -- Raid is disbanded, log everyone as having left and end the raid
	 mod:StartOrEndRaid()
      else
	 local name
	 -- Log everyone that joined since last scan
	 for id = 1, num do
	    name = GetRaidRosterInfo(id)
	    if name then
	       if not raidMembers[name] then
		  mod:LogPlayerJoin(name)
	       end
	       newRaidMembers[name] = true
	    end
	 end
	 -- Log anyone that left since last scan
	 for name in pairs(raidMembers) do
	    if not newRaidMembers[name] then
	       mod:LogPlayerLeave(name)
	    end
	 end
	 -- Swap lists
	 raidMembers, newRaidMembers = newRaidMembers, raidMembers 
      end
      for name in pairs(newRaidMembers) do newRaidMembers[name] = nil end
   end
end

function mod:StartOrEndRaid()
   local num = GetNumRaidMembers()
   if currentRaid and  currentRaid.endTime then
      currentRaid = nil
      currentEvent = nil
   end

   if num >= tonumber(db.raidsize) then
      if mod:IsLootMaster() then
	 -- We're the raid leader / loot master, so start a new raid
	 if not currentRaid then
	    if self.hasTrace then self:trace("No current raid - creating new or finding active. ") end
	    self:FindOrCreateRaid(true)
	 else
	    mod:CheckRaidChanges(true)
	 end
	 self:RegisterEvent("UPDATE_MOUSEOVER_UNIT", "ScanTarget", "mouseover")   
	 self:RegisterEvent("PLAYER_TARGET_CHANGED", "ScanTarget", "target")   
	 if db.lootEventListen then
	    self:RegisterEvent("CHAT_MSG_LOOT")
	 end
	 self:RegisterEvent("PLAYER_REGEN_DISABLED")
	 self:RegisterEvent("RAID_ROSTER_UPDATE", "CheckRaidChanges")
      else
	 if mod.masterNotRequested then
	    mod.masterNotRequested = true
	    mod:RequestInitialSyncData()
	    if self.hasTrace then mod:trace("Slave mode enabled, requesting raid info from the raid master.") end
	 end
      end
   elseif num == 0 then
      mod.masterNotRequested = nil

      if not currentRaid then
	 currentRaid = raids[#raids]
      end
      if currentRaid and not currentRaid.endTime then
	 if self.hasTrace then self:trace("Current raid should be ended.") end
	 self:EndRaid() -- we're no longer in a raid at all, so end the raid
      end
      self:UnregisterEvent("PLAYER_REGEN_DISABLED")
      self:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")   
      self:UnregisterEvent("PLAYER_TARGET_CHANGED")
      if db.lootEventListen then
	 self:UnregisterEvent("CHAT_MSG_LOOT")
      end
   end   
end

function mod:LogPlayerLeave(name)
   if mod.hasTrace then mod:trace("%s has left the raid.", name) end
   if not db.logPlayerEventData then return end
   if not currentRaid then return end

   local pe = currentRaid.playerEvents or {}

   if not pe[name] then
      pe[name] = { { 0, time() } }
   else
      pe[name][#pe[name]][2] = time() 
   end
   currentRaid.playerEvents = pe
end

function mod:LogPlayerJoin(name)
   if mod.hasTrace then mod:trace("%s has joined the raid", name) end
   mod:SendDKPBalance(name) -- send their current DKP balance
   if not db.logPlayerEventData then return end
   if not currentRaid then return end
   local pe = currentRaid.playerEvents or {}
   if not pe[name] then pe[name] = { } end
   local cnt = #pe[name]
   if cnt == 0 or pe[name][cnt][2] ~= 0 then
      pe[name][cnt+1] = {
	 time(), 0
      }
   end
   currentRaid.playerEvents = pe
end


function mod:GetGameTimeAsUnix()
   -- This does NOT work - if local time just swapped across midnight, but gametime is nto we get a 24 hr error.
   --   local h, m = GetGameTime()
   --   local n = date("*t")
   --   n.hour = h
   --   n.min = m
   --   return time(n)
   -- 
   return time() 
end

function mod:FromUnixTime(timestamp, prefix, postfix, short)
   local dt = date("*t", timestamp)
   if short then
      return fmt("%s%02d/%02d %s", prefix or "" , dt.month, dt.day, postfix or "")
   else
      return fmt("%s%02d/%02d/%d %02d:%02d:%02d%s", prefix or "" , dt.month, dt.day, dt.year, dt.hour, dt.min, dt.sec, postfix or "")
   end
end

function mod:LogCustomEvent(eventid)
   if not currentRaid then
      if self.error then self:error("No active raid - not logging event '%s'", events[eventid]) end
      return
   end
   local oldevent = currentEvent
   currentEvent =  currentRaid.events[events[eventid]]
   if not currentEvent then
      currentEvent = {
	 attendees = {},
	 standby = {},
	 name = events[eventid],
	 zone = currentZone.name,
      }
   end

   if not currentEvent.eventTime or not next(currentEvent.attendees) then
      if self.hasInfo then self:info("Taking Attendance for: %s", events[eventid]) end
      self:TakeAttendance()
   elseif self.hasInfo then
      self:info("Event %s already logged with attendance.", events[eventid])
   end
   currentEvent = oldevent
end

function mod:AddWipe()
   if not currentRaid then
      if self.error then self:error("No active raid - not logging wipe.") end
      return
   elseif not currentBoss then
      if self.error then self:error("No boss recently engaged - not logging wipe.") end
      return
   end

   
   
   local event = fmt("%s - %s #", L["Wipe"],
			       currentZone.mobs[currentBoss].event or
				  BL[currentZone.mobs[currentBoss].name] or
				  currentZone.mobs[currentBoss].name)

   local eventnum
   currentBoss = nil
   eventInProgress = nil
   for id = 1, 100 do
      eventnum = event..id
      if not currentRaid.events[eventnum] then
	 currentEvent = {
	    attendees = {},
	    standby = {},
	    zone = currentZone.name,
	    name = eventnum,
	 }
	 currentRaid.events[eventnum] = currentEvent
	 if self.hasInfo then self:info("Logging Wipe: %s", eventnum) end
	 self:TakeAttendance()
	 currentEvent = nil
	 return
      end
   end
end

function mod:OnProfileChanged(event, newdb)
   if event ~= "OnProfileDeleted" then
      db = self.db.profile
      self:SetLogLevel(db.logLevel)
   end
      
   self:NotifyChange()

end


local exportDialog
-- Ugly export function for an ugly XML format, written in an ugly way
-- to work with a stupid importer. Yeah this is some messy stuff.

function mod:CreateExportDialog()
   exportDialog = GUI:Create("MDKP_Dialog")
   exportDialog:SetTitle(L["DKP Export"])
   exportDialog:SetLayout("Fill")
   exportDialog.editbox:SetWidth(350)
   exportDialog.editbox:SetNumeric(false)
   exportDialog:SetText(L["Click on the text field below and copy the data to export it."])
   
   exportDialog.close = GUI:Create("Button")
   
   exportDialog:AddChild(exportDialog.close)
   exportDialog.close:SetText(L["Close"])
   
   exportDialog.close:SetCallback("OnClick",
				  function()
				     exportDialog:Hide()
				     exportDialog.editbox:SetText("")
				  end
			       )
   mod.CreateExportDialog = nil
end

function mod:ExportCTRT(info)
   local itempat = '|c(........)|Hitem:(.-)|h%['
   local raidid = tonumber(type(info) == "table" and info[#info-1] or info)
   if raids[raidid] then
       local dummyItemId = 21877 -- netherweave
      if not GetItemInfo(dummyItemId) then
	 dummyItemId = 33470 -- frostweave
      end
      if not GetItemInfo(dummyItemId) then
	 dummyItemId = 33447 -- runic healing potion
      end
      local itemdummy = {
	 {
	    value = 0,
	    looter = "disenchant",
	    itemid = dummyItemId,
	    count = 1
	 }
      }
      local raid = raids[raidid]
      local loot = ""
      local lootcount, count = 1, 1
      local zone = raid.zone or "Unknown"
      local export = "<RaidInfo>"
	 .."<Version>"..mod.MAJOR_VERSION.."."..mod.MINOR_VERSION.."</Version>"
	 .. self:FromUnixTime(raid.startTime, "<key>", "</key>")
	 .. self:FromUnixTime(raid.startTime, "<start>", "</start>")
	 .. self:FromUnixTime(raid.endTime or raid.startTime , "<end>", "</end>")
	 .. "<zone>"..zone.."</zone>"

      local bosskills = ""

      local jointxt, leavetxt = "", ""
      local joincount = 1
      local sortevents = {}
      for id in pairs(raid.events) do
	 sortevents[#sortevents+1] = id
      end
      table.sort(sortevents,
		 function(a, b)
		    return (raid.events[a].eventTime or 0) < (raid.events[b].eventTime or 0)
		 end)
      local alts = {}
      for id,data in pairs(players) do
	 if data.main then
	    alts[id] = data.main
	 end
      end
      local addevent = function(data, id)
			  local attendees = ""
			  local acount = 1
			  local eventTime = data.eventTime or raid.startTime
			  local eventTimeStr = self:FromUnixTime(eventTime)
			  
			  local attendee = function(name)
					      local pinfo
					      name = alts[name] or name
					      pinfo = players[name]
					      attendees = fmt("%s<key%d><name>%s</name></key%d>", attendees, acount, name, acount)
					      if pinfo then
						 if pinfo.sex and badData[pinfo.sex] then pinfo.sex = nil end
						 if pinfo.race and badData[pinfo.race] then pinfo.race = nil end
					      end
					      jointxt = fmt("%s<key%d><player>%s</player><race>%s</race><class>%s</class><sex>%d</sex><level>%d</level><time>%s</time></key%d>", jointxt, joincount, name,
								      pinfo and pinfo.race or "", pinfo and pinfo.class or "", tonumber(pinfo and pinfo.sex) or 0, pinfo and pinfo.level or 0, 
								      self:FromUnixTime(eventTime-1), joincount)
					      leavetxt = fmt("%s<key%d><player>%s</player><time>%s</time></key%d>", leavetxt, joincount, name, self:FromUnixTime(eventTime+1), joincount)
					      acount = acount + 1
					      joincount = joincount + 1
					   end
			  
			  for name in pairs(data.attendees) do
			     attendee(name)
			  end
			  
			  if db.standbyattendee and data.standby then
			     for name in pairs(data.standby) do
				attendee(name)
			     end
			  end
			  
			  bosskills = fmt("%s<key%d><name>%s</name><time>%s</time><attendees>%s</attendees></key%d>",
						    bosskills, count, id, eventTimeStr, attendees, count)
			  count = count + 1
			  local tmpitems
			  local overridezone
			  if data.items and next(data.items) then
			     tmpitems = data.items
			  else
			     tmpitems = itemdummy -- Stupid CTRT wont' import event without items
			     if db.eventnameaszone then
				for _, data in pairs(events) do
				   if id == data then
				      overridezone = id
				      break
				   end
				end
			     end
			  end
			  
			  for _,item in pairs(tmpitems) do
			     local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture
			     if items[item.itemid] then
				itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType,
				itemSubType, itemStackCount, itemEquipLoc, itemTexture =  unpack(items[item.itemid].itemInfo)
			     end
			     
			     if not itemName then 
				itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType,
				itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(item.itemid)
			     end
			     
			     --|cffffffff|Hitem:23984:0:0:0:0:0:0:0|h[Irradiated Crystal Shard]|h|
			     if self.hasSpam then self:spam("Exporting item %d : %s => %s", item.itemid, tostring(itemName), tostring(itemLink)) end
			     
			     local color, link = match(itemLink, itempat)
			     loot = fmt("%s<key%d><ItemName>%s</ItemName><ItemID>%s</ItemID><Icon>%s</Icon><Class>%s</Class><SubClass>%s</SubClass><Color>%s</Color><Count>%d</Count><Player>%s</Player><Costs>%d</Costs><Time>%s</Time><Zone>%s</Zone><Boss>%s</Boss><Note><![CDATA[ - Zone: %s - Boss: %s - %d DKP]]></Note></key%d>",
						  loot, lootcount, itemName, link, itemTexture, itemType, itemSubType,color,
						  item.count, alts[item.looter] or item.looter, item.value or 0, eventTimeStr,
						  overridezone or data.zone or zone, id, overridezone or data.zone or zone,
						  id, item.value or 0, lootcount)
			     lootcount = lootcount + 1
			  end
		       end
      local cache = {}
      for _,id in pairs(sortevents) do
	 local data = raid.events[id]
	 addevent(data, id)
      end
      export = fmt("%s<BossKills>%s</BossKills><Join>%s</Join><Leave>%s</Leave><Loot>%s</Loot></RaidInfo>", export, bosskills, jointxt, leavetxt, loot)
      if not exportDialog then mod:CreateExportDialog() end
      exportDialog.editbox:SetText(export)
      if C.OpenFrames["Magic DKP"] then
	 exportDialog.frame:SetFrameLevel(C.OpenFrames["Magic DKP"].frame:GetFrameLevel()+100)
      end
      exportDialog:Show()
   end
end

function mod:ExportMagicDKP(info)
   local raidid = tonumber(type(info) == "table" and info[#info-1] or info)
   local raid = raids[raidid]
   if not raid then return end
   local zone = raid.zone or "Unknown"
   local tinsert = table.insert
   local export = { }
   local raidPlayers = {}
   -- helperfunction
   local function ti(str, ...)
      if select("#", ...) > 0 then
	 tinsert(export, str:format(...))
      else
	 tinsert(export, str)
      end
   end

   local sortevents = {}
   for id in pairs(raid.events) do
      sortevents[#sortevents+1] = id
   end
   table.sort(sortevents,
	      function(a, b)
		 return (raid.events[a].eventTime or 0) < (raid.events[b].eventTime or 0)
	      end)

   local alts = {}
   for id,data in pairs(players) do
      if data.main then
	 alts[id] = data.main
      end
   end
   ti('<?xml version="1.0" encoding="UTF-8"?>')
   ti("<raid>")
   ti("  <starttime>%d</starttime>", raid.startTime)
   ti("  <endtime>%d</endtime>", raid.endTime or raid.startTime)
   ti("  <zone>%s</zone>", zone)
   ti("  <events>")
   local att = {}
   for _,id in ipairs(sortevents) do
      for id in pairs(att) do att[id] = nil end
      local data = raid.events[id]
      ti("    <event>")
      ti("       <name>%s</name>", data.name or data.zone or zone)
      ti("       <zone>%s</zone>", data.zone or zone)
      ti("       <time>%d</time>", data.eventTime or raid.startTime)
      if next(data.attendees) then 
	 ti("       <attendees>")
	 for name in pairs(data.attendees) do
	    local n = alts[name] or name
	    att[n] = true
	    raidPlayers[n] = true
	    ti("         <player>%s</player>", n)
	 end
	 ti("       </attendees>")
      end
      local c = ITEM_QUALITY_COLORS
      if data.standby and next(data.standby) then
	 ti("       <standby>")
	 for name in pairs(data.standby) do
	    local n = alts[name] or name
	    if not att[n] then
	       raidPlayers[n] = true
	       ti("         <player>%s</player>", n)
	    end
	 end
	 ti("       </standby>")
      end

      if data.items and next(data.items) then
 	 ti("       <items>")
	 for _,item in pairs(data.items) do
	    local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture
	    if items[item.itemid] then
	       itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType,
	       itemSubType, itemStackCount, itemEquipLoc, itemTexture =  unpack(items[item.itemid].itemInfo)
	    end	    
	    if not itemName then 
	       itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType,
	       itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo(item.itemid)
	    end
	    ti("         <item>")
	    ti("           <name>%s</name>", itemName)
	    ti("           <itemid>%d</itemid>", mod:LinkToItemID(itemLink))
	    ti("           <player>%s</player>", alts[item.looter] or item.looter)
	    ti("           <cost>%d</cost>", item.value or 0)
	    ti("           <count>%d</count>", item.count)
	    ti("           <itemcolor>%02x%02x%02x</itemcolor>", c[itemRarity].r*255, c[itemRarity].g*255, c[itemRarity].b*255)
	    ti("           <itemlevel>%s</itemlevel>", itemLevel)
	    ti("           <rarity>%d</rarity>", itemRarity)
	    ti("           <minlevel>%s</minlevel>", itemMinLevel)
	    ti("           <type>%s</type>", itemType)
	    ti("           <subtype>%s</subtype>", itemSubType)
	    ti("           <equiploc>%s</equiploc>", itemEquipLoc)
	    ti("           <icon>%s</icon>", itemTexture)
	    ti("         </item>")
	 end
 	 ti("       </items>")
      end
      ti("    </event>")
      
   end
   ti("  </events>")
   ti("  <players>")

   -- ensures we got all the event (join/leave) data, merging alts if necessary   
   if raid.playerEvents then
      for name, data in pairs(raid.playerEvents) do
	 local n = alts[name] or name
	 if type(raidPlayers[n]) ~= "table" then
	    raidPlayers[n] = {}
	 end
	 for _,event in pairs(data) do
	    tinsert(raidPlayers[n], event)
	 end
	 table.sort(raidPlayers[n], function(a, b) return a[1] > a[2] end)
      end
   end
   for name, data in pairs(raidPlayers) do
      local pi = players[name]
      ti("    <player>")
      ti("      <name>%s</name>", name)
      if pi then 
	 ti("      <class>%s</class>", pi.class)
	 ti("      <race>%s</race>", pi.race)
	 ti("      <level>%d</level>", pi.level)
	 ti("      <rank>%s</rank>", pi.rank)
	 ti("      <sex>%s</sex>", pi.sex)
      end
      if type(data) == "table" then 
	 local seen = {}
	 ti("      <raidpresence>")
	 for _,event in ipairs(data) do
	    if not seen[event[1]] then
	       seen[event[1]] = true
	       ti("        <period>")
	       if event[1] > 0 then ti("          <start>%d</start>", event[1]) end
	       if event[2] > 0 then ti("          <end>%d</end>", event[2])     end
	       ti("        </period>")	 
	    end
	 end
	 ti("      </raidpresence>")
      end
      ti("    </player>")
   end
   ti("  </players>")
   ti("</raid>")
   if not exportDialog then mod:CreateExportDialog() end
   exportDialog.editbox:SetText(table.concat(export, "\n"))
   if C.OpenFrames["Magic DKP"] then
      exportDialog.frame:SetFrameLevel(C.OpenFrames["Magic DKP"].frame:GetFrameLevel()+100)
   end
   exportDialog:Show()
end

function mod:SendUrgentMessage(channel, recipient)
   MagicComm:SendUrgentMessage(networkData, "MD", channel, recipient)
end

function mod:SendBulkMessage(channel, recipient)
   if GetNumRaidMembers() > 0 then
      MagicComm:SendBulkMessage(networkData, "MD", channel, recipient)
   end
end

do
   local verRespMsg = "%s: %s revision %s"
   function mod:OnVersionResponse(ver, major, minor, sender)
      self:Print(verRespMsg:format(sender, major or "Unknown", minor or "Unknown"))
   end
end

function mod:QueryAddonVersions()
   mod:SetNetworkData("VCHECK")
   self:SendUrgentMessage("GUILD")
   self:SendUrgentMessage("RAID")
end

function mod:tokenize(str, values)
   for k, v in pairs(values) do
      str = gsub(str, "%["..k.."%]", (type(v) == "function" and v() or v))
   end
   return str
end

function mod:GetModuleDatabase(module, default, options)
   local optionPane
   defaultOptions.profile.modules[module] = default
   self.db:RegisterDefaults(defaultOptions)
   if options then
      optionPane = mod:RegisterModuleOptions(module, options)
   end
   return db.modules[module], optionPane
end

function mod:SendChatMessage(message, destination, recipient, prefix)
   if destination == "RW" then
      destination = (IsRaidLeader() or IsRaidOfficer()) and "RAID_WARNING" or "GROUP"
   end
   if destination == "GROUP" then
      destination = GetNumRaidMembers() > 0 and "RAID" or "PARTY"
   end
   if prefix then
      message = prefix..message
   end
   SendChatMessage(message, destination, nil, recipient)
end

do 
   local msg = {}
   function mod:SendDKPBalance(player)
      if not UnitExists(player) then return end 
      local DB = mod:Database()
      if not DB then return end
      local data = DB:GetPlayerData(player)
      if not data and players[player] and players[player].main then
	 data = DB:GetPlayerData(players[player].main)
      end
      if data then
	 for id in pairs(msg) do msg[id] = nil end
	 for id, data in pairs(data) do
	    if id ~= "items" and id ~= "class" then
	       msg[id] = data
	    end
	 end
	 mod:SetNetworkData("DKPBALANCE", msg)
	 mod:SendUrgentMessage("WHISPER", recipient or player)
      end
   end
end

function mod:GetMultiPool(item)
   local DB = mod:Database()

   if type(item) == "string" then item = mod:LinkToItemID(item) end
   local pools = DB:GetMultiDKPPools() 

   if item and items[item] then
      local md = items[item].multiDKP
      if md and pools[md] then
	 return md  -- item specific override
      end
   end
   
   if currentZone and currentZone.multiDKP and pools[currentZone.multiDKP] then
      -- Zone wide multi dkp setting
      return currentZone.multiDKP
   end
   
   if DB then 
      return db.modules.Database.defaultMulti or "dkp"
   end
end

function mod:GetMultiPools()
   local DB = mod:Database()
   if DB then return DB:GetMultiDKPPools() end
end

function mod:GetPlayerMain(player)
   return players[player] and players[player].main
end

function mod:Database()
   return mod.modules.Database
end

function mod:NoDKPDatabase()
   return mod.modules.Database == nil
end

function mod:LinkToItemID(link)
   return tonumber(match(link, ".*|Hitem:(%d+):"))
end
